"""Factiva Core OAuthUser Class."""
from ..tools import load_environment_value


class OAuthUser:
    """Class that represents an API user. This entity is identifiable by a User Key.

    Parameters
    ----------
    key : str
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    stats : bool, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.

    """

    client_id = ''
    username = ''
    password = ''


    def __init__(
        self,
        client_id=None,
        username=None,
        password=None
    ):
        """Constructs the instance of the class."""
        if client_id is None:
            try:
                self.client_id = load_environment_value('FACTIVA_CLIENTID')
            except Exception as error:
                raise ValueError('Factiva client ID not provided and environment variable FACTIVA_CLIENTID not set.') from error

        if username is None:
            try:
                self.username = load_environment_value('FACTIVA_USERNAME')
            except Exception as error:
                raise ValueError('Factiva username not provided and environment variable FACTIVA_USERNAME not set.') from error

        if password is None:
            try:
                self.password = load_environment_value('FACTIVA_PASSWORD')
            except Exception as error:
                raise ValueError('Factiva password not provided and environment variable FACTIVA_PASSWORD not set.') from error


    def get_auth_token(self):
        """To be implemented"""


    def get_jwt_token(self):
        """To be implemented"""


    def __print_property__(self, property_value) -> str:
        if isinstance(property_value, int):
            pval = f'{property_value:,d}'
        else:
            pval = property_value
        return pval


    def __repr__(self):
        """Return a string representation of the object."""
        return self.__str__()


    def __str__(self, detailed=True, prefix='  |-', root_prefix=''):
        pprop = self.__dict__.copy()

        ret_val = f'{root_prefix}{str(self.__class__)}\n'
        if detailed:
            ret_val += '\n'.join((f'{prefix}{item} = {self.__print_property__(pprop[item])}' for item in pprop))
        else:
            ret_val += f'{prefix}...'
        return ret_val



"""
    Define methods and properties for tools module.
"""
__all__ = [
    'snapshot', 'stream', 'bulknews', 'taxonomy', 'Snapshot', 'SnapshotQuery',
     'Stream', 'Listener', 'Subscription', 'Taxonomy', 'tools',
    'SnapshotFiles', 'factiva_logger', 'get_factiva_logger', # 'JSONLFileHandler', 'BigQueryHandler', 'MongoDBHandler'
    'UserKey', 'factiva_logger', 'get_factiva_logger', 'StreamResponse'
]

from .__version__ import __version__
from .auth.userkey import UserKey
from .tools.files import SnapshotFiles
from .log import factiva_logger, get_factiva_logger
from .snapshot import Snapshot, SnapshotQuery
from .stream import Listener, Stream, Subscription
from .taxonomy.taxonomy import Taxonomy
from .stream_response import StreamResponse
# from .tools import JSONLFileHandler, BigQueryHandler, MongoDBHandler

version = __version__



import copy
import datetime
import json
import os

from factiva.core import const, factiva_logger, get_factiva_logger, tools
from google.cloud import bigquery
from pymongo import MongoClient

from .bq_schemas import *

class JSONLFileHandler:
    def __init__(self):
        """Initialize class constructor."""
        self.counter = 0
        self.log = get_factiva_logger()

    def write_jsonl_line(self, file_prefix, action, file_suffix, message):
        """Write a new Jsonl line.
        
        Parameters
        ----------
        file_prefix : str
            File prefix
        action : str
            Action from the stream response
        file_suffix : str
            File suffix
        message : str
            Message to be write on the file
        """
        output_filename = f'{file_prefix}_{action}_{file_suffix}.jsonl'
        output_filepath = os.path.join(const.LISTENER_FILES_DEFAULT_FOLDER,
                                       output_filename)
        with open(output_filepath, mode='a', encoding='utf-8') as fp:
            fp.write(
                f"{json.dumps(message, ensure_ascii=False, sort_keys=True, default=str)}\n")

    @factiva_logger
    def save(self, message, subscription_id) -> bool:
        """Listener to save response into a jsonl
        
        Parameters
        ----------
        message : str
            Message to be stored
        subscription_id : str
            Suscription id from response
        
        Returns
        -------
        bool:
            Status from the process
        """
        self.log.info("Saving into JSONL file")
        tools.create_path_if_not_exist(const.LISTENER_FILES_DEFAULT_FOLDER)

        errorFile = os.path.join(const.LISTENER_FILES_DEFAULT_FOLDER, 'errors.log')
        erroMessage = f"{datetime.datetime.utcnow()}\tERR\t$$ERROR$$\t$$MESSAGE$$\n"

        stream_short_id = subscription_id.split('-')[-3]
        current_hour = datetime.datetime.utcnow().strftime('%Y%m%d%H')

        if 'action' in message.keys():

            message = tools.format_timestamps(message)
            message = tools.format_multivalues(message)
            current_action = message['action']

            if current_action in const.ALLOWED_ACTIONS:
                print(const.ACTION_CONSOLE_INDICATOR[current_action], end='')
                self.write_jsonl_line(stream_short_id, current_action,
                                      current_hour, message)
            else:
                print(const.ACTION_CONSOLE_INDICATOR[const.ERR_ACTION], end='')
                with open(os.path.join(errorFile), mode='a',
                          encoding='utf-8') as efp:
                    efp.write(
                        erroMessage.replace('$$ERROR$$',
                                            'InvalidAction').replace(
                                                '$$MESSAGE$$',
                                                json.dumps(message)))
            self.counter += 1
            if self.counter % 100 == 0:
                print(f'\n[{self.counter}]', end='')

        else:
            print(const.ACTION_CONSOLE_INDICATOR[const.ERR_ACTION], end='')
            with open(errorFile, mode='a', encoding='utf-8') as efp:
                efp.write(
                    erroMessage.replace('$$ERROR$$', 'InvalidMessage').replace(
                        '$$MESSAGE$$', json.dumps(message)))
            return False
        return True

class BigQueryHandler:
    def __init__(self):
        """Initialize class constructor."""
        self.table_id = os.getenv('STREAMLOG_BQ_TABLENAME', None)
        if self.table_id is None:
            raise RuntimeError('Env variable STREAMLOG_BQ_TABLENAME not set')
        
        self.client = bigquery.Client()
        self.counter = 0
        self.log_line = ''
        self.log = get_factiva_logger()

    @factiva_logger
    def save(self, message, subscription_id) -> bool:
        """Listener to save response into a big query table
        
        Parameters
        ----------
        message : str
            Message to be stored
        subscription_id : str
            Suscription id from response
        
        Returns
        -------
        bool:
            Status from the process
        """
        self.log.info("Saving into BigQuery table")
        tools.create_path_if_not_exist(const.LISTENER_FILES_DEFAULT_FOLDER)
        errorFile = os.path.join(const.LISTENER_FILES_DEFAULT_FOLDER, 'errors.log')
        erroMessage = f"{datetime.datetime.utcnow()}\tERR\t$$ERROR$$\t$$MESSAGE$$\n"
        ret_val = False
        _msg = copy.deepcopy(message)
        msg_an = _msg['an']

        try:
            if 'action' in _msg.keys():
                msg_an = _msg['an']
                current_action = _msg['action']
                if current_action in const.ALLOWED_ACTIONS:
                    _msg = tools.format_timestamps(_msg)
                    _msg = tools.format_multivalues(_msg)
                    _msg = format_message_to_response_schema(_msg)
                    errors = self.client.insert_rows_json(
                        self.table_id, [_msg])
                    self.log_line += const.ACTION_CONSOLE_INDICATOR[
                                current_action]
                    if errors == []:
                        ret_val = True
                else:
                    self.log_line += const.ACTION_CONSOLE_INDICATOR[
                        const.ERR_ACTION]
                
                self.counter += 1
                if self.counter % 100 == 0:
                    self.log_line+= '\n'
            else:
                print(const.ACTION_CONSOLE_INDICATOR[const.ERR_ACTION], end='')
                with open(errorFile, mode='a', encoding='utf-8') as efp:
                    efp.write(
                        erroMessage.replace('$$ERROR$$', 'InvalidMessage').replace(
                            '$$MESSAGE$$', json.dumps(message)))
            return False

        except Exception as e:
            log_path = const.LISTENER_FILES_DEFAULT_FOLDER
            if msg_an != '':
                file_name = os.path.join(log_path, f"{msg_an}.json")
            else:
                file_name = os.path.join(log_path,
                                         f"{tools.now_to_tssec()}.json")

            with open(file_name, mode='a', encoding='utf-8') as fp:
                fp.write(
                    f"{json.dumps(message, ensure_ascii=False, sort_keys=True)}\n"
                )
            ret_val = True
            self.log_line += '#'
            self.counter += 1
        print(self.log_line)
        return ret_val

    def close_connection(self):
        self.client.close()

class MongoDBHandler:

    def __init__(self):
        """Initialize class constructor."""
        self.counter = 0
        self.log = get_factiva_logger()

        connection_string = os.getenv('MONGODB_CONNECTION_STRING', None)
        database_name = os.getenv('MONGODB_DATABASE_NAME', None)
        collection_name = os.getenv('MONGODB_COLLECTION_NAME', None)
        if connection_string is None or database_name is None or collection_name is None:
            self.log.error('MongoDB environment vars are not set')
            raise RuntimeError('MongoDB environment vars are not set')

        self.client = MongoClient(connection_string)
        self.database = self.client[database_name]
        self.mongodb_collection = self.database[collection_name]


    @factiva_logger
    def save(self, message, subscription_id) -> bool:
        """Listener to save response into a mongodb table
        
        Parameters
        ----------
        message : str
            Message to be stored
        subscription_id : str
            Suscription id from response
        
        Returns
        -------
        bool:
            Status from the process
        """
        self.log.info("Saving into MongoDB")

        tools.create_path_if_not_exist(const.LISTENER_FILES_DEFAULT_FOLDER)

        errorFile = os.path.join(const.LISTENER_FILES_DEFAULT_FOLDER, 'errors.log')
        erroMessage = f"{datetime.datetime.utcnow()}\tERR\t$$ERROR$$\t$$MESSAGE$$\n"

        if 'action' in message.keys():

            message = tools.format_timestamps_mongodb(message)
            message = tools.format_multivalues(message)
            current_action = message['action']

            if current_action in const.ALLOWED_ACTIONS:
                print(const.ACTION_CONSOLE_INDICATOR[current_action], end='')
                self.mongodb_collection.insert_one(message)

            else:
                print(const.ACTION_CONSOLE_INDICATOR[const.ERR_ACTION], end='')
                with open(os.path.join(errorFile), mode='a',
                          encoding='utf-8') as efp:
                    efp.write(
                        erroMessage.replace('$$ERROR$$',
                                            'InvalidAction').replace(
                                                '$$MESSAGE$$',
                                                json.dumps(message, default=str)))

            self.counter += 1
            if self.counter % 100 == 0:
                print(f'\n[{self.counter}]', end='')

        else:
            print(const.ACTION_CONSOLE_INDICATOR[const.ERR_ACTION], end='')
            with open(errorFile, mode='a', encoding='utf-8') as efp:
                efp.write(
                    erroMessage.replace('$$ERROR$$', 'InvalidMessage').replace(
                        '$$MESSAGE$$', json.dumps(message, default=str)))
            return False
        return True

    def close_connection(self):
        self.client.close()



"""Multiple methods and other resources for the Factiva Analytics package."""
# __all__ = ['JSONLFileHandler', 'BigQueryHandler', 'MongoDBHandler']

# from .listener_handlers import JSONLFileHandler, BigQueryHandler, MongoDBHandler

import datetime
import os
from dateutil import parser
from .. import const


def load_environment_value(config_key, default=None) -> str:
    """Obtain a environmental variable."""
    tmp_val = os.getenv(config_key, default)
    if tmp_val is None:
        raise Exception(
            "Environment Variable {} not found!".format(config_key))
    return tmp_val


def mask_string(raw_str, right_padding=4) -> str:
    """Mask a string with a default value of 4."""
    return raw_str[-right_padding:].rjust(len(raw_str), '*')


def validate_type(var_to_validate, expected_type, error_message) -> bool:
    """Validate a given type."""
    if not isinstance(var_to_validate, expected_type):
        raise ValueError(error_message)


def validate_field_options(field, available_options):
    """Validate that field is among the available options.
    Parameters
    ----------
    field:
        field to be validated, could be any type.
    available_options:
        options amongst field should be located.
    """
    if field not in available_options:
        raise ValueError(
            f'Value {field} is not within the allowed options: {available_options}'
        )


def flatten_dict(multi_level_dict) -> dict:
    """Flatten a dictionary."""
    flattened_dict = {}
    for entry in multi_level_dict:
        if isinstance(multi_level_dict[entry], dict):
            new_elements = flatten_dict(multi_level_dict[entry])
            flattened_dict.update(new_elements)
        else:
            flattened_dict[entry] = multi_level_dict[entry]

    return flattened_dict


def isots_to_tsms(isodatestr: str) -> int:
    return round(
        datetime.datetime.fromisoformat(
            str(isodatestr).replace('&', '')
        ).timestamp())


def now_to_tsms() -> int:
    return round(datetime.datetime.utcnow().timestamp())


def format_timestamps(message: dict) -> dict:
    """Format datetimes from a dict
    Parameters
    ----------
    message:
        dict with datetime values to be formated
    Returns
    -------
    dict
        Dict with datetimes formated
    """
    for fieldname in const.TIMESTAMP_FIELDS:
        if fieldname in message.keys():
            message[fieldname] = isots_to_tsms(message[fieldname])
    message["delivery_datetime"] = now_to_tsms()
    return message


def format_timestamps_mongodb(message: dict) -> dict:
    """Format datetimes into mongodb datetime from a dict
    Parameters
    ----------
    message:
        dict with datetime values to be formated
    Returns
    -------
    dict
        Dict with datetimes formated
    """
    for fieldname in const.TIMESTAMP_FIELDS:
        if fieldname in message.keys():
            message[fieldname] = parser.parse(message[fieldname])
    message["delivery_datetime"] = parser.parse(
        datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
    return message


def multivalue_to_list(fieldvalue: str, sep=',') -> list:
    if (fieldvalue is None) or (fieldvalue == ''):
        retval = []
    else:
        allvals = fieldvalue.split(sep)
        allvals = [x for x in allvals if x != '']
        retval = list(set(allvals))  # Removes duplicates
    return retval


def format_multivalues(message: dict) -> dict:
    """Format multivalues from a dict
    Parameters
    ----------
    message:
        dict with multivalues values to be formated
    Returns
    -------
    dict
        Dict with multivalues formated
    """
    for fieldname in const.MULTIVALUE_FIELDS_SPACE:
        if fieldname in message.keys():
            message[fieldname] = multivalue_to_list(message[fieldname],
                                                    sep=' ')
    for fieldname in const.MULTIVALUE_FIELDS_COMMA:
        if fieldname in message.keys():
            message[fieldname] = multivalue_to_list(message[fieldname])
    return message


def create_path_if_not_exist(path):
    """Check if a path exist, if not then is created
    Parameters
    ----------
    path:
        path to create
    Returns
    -------
    dict
        Dict with multivalues formated
    """
    if not os.path.exists(path):
        os.makedirs(path)



"""Implement query class definition."""
from .. import const
from ..tools import validate_type, validate_field_options
from factiva.analytics.bulknews import BulkNewsQuery


def validate_group_options(by_source, by_dimensions):
    if (isinstance(by_source, bool) and isinstance(by_dimensions, list)):
        raise ValueError(
            "The value group_by_source_code and group_dimensions are not compatible each other"
        )

class SnapshotQuery(BulkNewsQuery):
    """Implement Snapshot Query class definition."""
    limit = 0
    file_format = ''
    frequency = ''
    date_field = ''
    group_by_source_code = None
    top = 0
    group_dimensions = None

    # pylint: disable=too-many-arguments
    def __init__(self,
                 where,
                 includes=None,
                 excludes=None,
                 select_fields=None,
                 limit=0,
                 file_format=const.API_AVRO_FORMAT,
                 frequency=const.API_MONTH_PERIOD,
                 date_field=const.API_PUBLICATION_DATETIME_FIELD,
                 group_by_source_code=None,
                 group_dimensions=None,
                 top=10):
        """Initialize query class instance."""
        super().__init__(where=where,
                         includes=includes,
                         excludes=excludes,
                         select_fields=select_fields)

        validate_type(limit, int, "Unexpected value for limit")
        if limit >= 0:
            self.limit = limit
        else:
            raise ValueError("Limit value is not valid or not positive")

        validate_type(file_format, str, "Unexpected value for file_format")
        file_format = file_format.lower().strip()
        validate_field_options(file_format, const.API_EXTRACTION_FILE_FORMATS)
        self.file_format = file_format

        validate_type(frequency, str, "Unexpected value for frequency")
        frequency = frequency.upper().strip()
        validate_field_options(frequency, const.API_DATETIME_PERIODS)
        self.frequency = frequency

        validate_type(date_field, str, "Unexpected value for date_field")
        date_field = date_field.lower().strip()
        validate_field_options(date_field, const.API_DATETIME_FIELDS)
        self.date_field = date_field

        validate_group_options(group_by_source_code, group_dimensions)

        if isinstance(group_by_source_code, bool):
            self.group_by_source_code = group_by_source_code

        if isinstance(group_dimensions, list):
            self.group_dimensions = group_dimensions

        validate_type(top, int, "Unexpected value for top")
        if top >= 0:
            self.top = top
        else:
            raise ValueError('Top value is not valid')

    def get_explain_query(self):
        """Obtain Base Query."""
        query_dict = self.get_base_query()
        return query_dict

    def get_analytics_query(self):
        """Obtain analytics Query."""
        query_dict = self.get_base_query()

        self.frequency = self.frequency.upper().strip()
        validate_field_options(self.frequency, const.API_DATETIME_PERIODS)
        query_dict["query"].update({"frequency": self.frequency})

        self.date_field = self.date_field.lower().strip()
        validate_field_options(self.date_field, const.API_DATETIME_FIELDS)
        query_dict["query"].update({"date_field": self.date_field})

        validate_group_options(self.group_by_source_code, self.group_dimensions)

        if(isinstance(self.group_by_source_code, bool)):      
            query_dict["query"].update(
                {"group_by_source_code": self.group_by_source_code})
        else:
            if(self.group_dimensions):
                if(len(self.group_dimensions)<=4):
                    for option in self.group_dimensions:
                        validate_field_options(option, const.API_GROUP_DIMENSIONS_FIELDS)
                else:
                    raise ValueError("The maximiun group_dimensions options is 4")
            else:
                self.group_dimensions = []
            
            query_dict["query"].update(
                {"group_dimensions": self.group_dimensions})
        
        query_dict["query"].update({"top": self.top})
        return query_dict

    def get_extraction_query(self):
        """Obtain the string querying Factiva."""
        query_dict = self.get_base_query()

        if self.limit > 0:
            query_dict["query"].update({"limit": self.limit})

        query_dict["query"].update({"format": self.file_format})

        return query_dict

    def __repr__(self):
        """Create string representation for Query Class."""
        return self.__str__()

    def __str__(self, detailed=False, prefix='  |-', root_prefix=''):
        """Create string representation for Query Class."""
        pprop = self.__dict__.copy()

        ret_val = f'{root_prefix}{str(self.__class__)}\n'
        if detailed:
            ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        else:
            ret_val += f'{prefix}where: '
            ret_val += (self.where[:77] + '...') if len(self.where) > 80 else self.where
            ret_val += f'\n{prefix}...'
            del pprop['where']
            ret_val += '\n'.join(('{}{} = {}'.format(prefix, item, pprop[item]) for item in pprop))
        return ret_val



"""Taxonomy class implementation."""
from io import StringIO

import pandas as pd
from .. import (UserKey, const, factiva_logger, get_factiva_logger,
                          req)
from ..tools import validate_type


class Taxonomy():
    """Class that represents the taxonomy available within the Snapshots API.

    Parameters
    ----------
    user_key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.

    Examples
    --------
    Creating a taxonomy instance providing the user key
        >>> t = Taxonomy(u='abcd1234abcd1234abcd1234abcd1234')

    Creating a taxonomy instance with an existing UserKey instance
        >>> u = UserKey('abcd1234abcd1234abcd1234abcd1234')
        >>> t = Taxonomy(user_key=u)

    """

    categories = []
    identifiers = []


    def __init__(self, user_key=None):
        """Class initializer."""
        self.user_key = UserKey.create_user_key(user_key)
        self.categories = self.get_categories()
        self.identifiers = self.get_identifiers()
        self.log= get_factiva_logger()

    @factiva_logger
    def get_categories(self) -> list:
        """Request for a list of available taxonomy categories.

        Returns
        -------
        List of available taxonomy categories.

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        This method is called with in the __init__ method, so the categories can be accessed as is.
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.categories)
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        Calling the method on its own
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.get_categories())
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        """
        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_TAXONOMY_BASEPATH}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            return [entry['attributes']['name'] for entry in response.json()['data']]

        raise RuntimeError('API Request returned an unexpected HTTP status')

    @factiva_logger
    def get_identifiers(self) -> list:
        """Request for a list of available taxonomy categories.

        Returns
        -------
        List of available taxonomy categories.

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        This method is called with in the __init__ method, so the categories can be accessed as is.
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.categories)
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        Calling the method on its own
            >>> taxonomy = Taxonomy()
            >>> print(taxonomy.get_categories())
            ['news_subjects', 'regions', 'companies', 'industries', 'executives']

        """
        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_COMPANY_IDENTIFIERS_BASEPATH}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            return response.json()['data']['attributes']

        raise RuntimeError('API Request returned an unexpected HTTP status')


    # TODO: Implement a save_path parameter to alternatively store the taxonomy
    # file locally, and don't load its content to a DataFrame
    # https://github.com/dowjones/factiva-news-python/issues/4#issue-956942535
    # Check also differences by loading the data in AVRO. In case the issue is
    # too common with Executives, force the download option.
    @factiva_logger
    def get_category_codes(self, category) -> pd.DataFrame:
        """Request for available codes in the taxonomy for the specified category.

        Parameters
        ----------
        category : str
            String with the name of the taxonomy category to request the codes from

        Returns
        -------
        Dataframe containing the codes for the specified category

        Raises
        ------
        ValueError: When category is not of a valid type
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Getting the codes for the 'industries' category
            >>> taxonomy = Taxonomy()
            >>> industry_codes = taxonomy.get_category_codes('industries')
            >>> print(industry_codes)
                    code                description
            0     i25121             Petrochemicals
            1     i14001         Petroleum Refining
            2       i257            Pharmaceuticals
            3     iphrws  Pharmaceuticals Wholesale
            4       i643     Pharmacies/Drug Stores

        """
        validate_type(category, str, 'Unexpected value: category value must be string')

        response_format = 'csv'

        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_TAXONOMY_BASEPATH}/{category}/{response_format}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict, stream=True)
        if response.status_code == 200:
            r_df = pd.read_csv(StringIO(response.content.decode()))
            if 'executiveFactivaCode' in r_df.columns:
                r_df.rename(columns = {'executiveFactivaCode':'code'}, inplace = True)
            elif 'Code' in r_df.columns:
                r_df.rename(columns = {'Code':'code'}, inplace = True)
            r_df.set_index('code', inplace=True)
            return r_df

        raise RuntimeError('API Request returned an unexpected HTTP Status')

    @factiva_logger
    def get_single_company(self, code_type, company_code) -> pd.DataFrame:
        """Request information about a single company.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        company_code : str
            String containing the company code

        Returns
        -------
        DataFrame containing the company information

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Get the company data using the code type 'isin' and the company code 'ABCNMST00394'
            >>> taxonomy = Taxonomy()
            >>> company_data = taxonomy.get_single_company('isin', 'ABCNMST00394')
            >>> print(company_data)
                         id  fcode           common_name
            0  ABCNMST00394  ABCYT  Systemy Company S.A.

        """
        validate_type(code_type, str, 'Unexpected value: code_type must be str')
        validate_type(company_code, str, 'Unexpected value: company must be str')

        headers_dict = {
            'user-key': self.user_key.key
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_COMPANIES_BASEPATH}/{code_type}/{company_code}'

        response = req.api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code == 200:
            response_data = response.json()
            return pd.DataFrame.from_records([response_data['data']['attributes']])

        raise RuntimeError('API Request returned an unexpected HTTP status')

    @factiva_logger
    def get_multiple_companies(self, code_type, company_codes) -> pd.DataFrame:
        """Request information about a list of companies.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        companies_codes : list
            List containing the company codes to request information about

        Returns
        -------
        DataFrame containing the company information

        Raises
        ------
        RuntimeError: When API request returns unexpected error

        Examples
        --------
        Get multiple companies data using the code type 'isin' and a company codes list
            >>> taxonomy = Taxonomy()
            >>> companies_data = taxonomy.get_multiple_companies('isin', ['ABC3E53433100', 'XYZ233341067', 'MN943181045'])
            >>> print(companies_data)
                         id   fcode      common_name
            0  ABC3E5343310  MCABST  M**************
            1  XYZ233341067   AXYZC    A************
            2  MN9431810453     MNM     M***********

        """
        validate_type(code_type, str, 'Unexpected value: code_type must be str')
        validate_type(company_codes, list, 'Unexpected value: companies must be list')
        for single_company_code in company_codes:
            validate_type(single_company_code, str, 'Unexpected value: each company in companies must be str')

        headers_dict = {
            'user-key': self.user_key.key
        }

        payload_dict = {
            "data": {
                "attributes": {
                    "ids": company_codes
                }
            }
        }

        endpoint = f'{const.API_HOST}{const.API_SNAPSHOTS_COMPANIES_BASEPATH}/{code_type}'

        response = req.api_send_request(method='POST', endpoint_url=endpoint, headers=headers_dict, payload=payload_dict)

        if response.status_code == 200 or response.status_code == 207:
            response_data = response.json()
            return pd.DataFrame.from_records(response_data['data']['attributes']['successes'])
        raise RuntimeError(f'API Request returned an unexpected HTTP status with message: {response.text}')

    @factiva_logger
    def get_company(self, code_type, company_codes) -> pd.DataFrame:
        """Request information about either a single company or a list of companies.

        Parameters
        ----------
        code_type : str
            String describing the code type used to request the information about the company. E.g. isin, ticker.
        company_code: str or list
            Single company code (str) or list of company codes to translate.

        Returns
        -------
        Dataframe with the information about the requested company(ies)

        Raises
        ------
        ValueError: When any given argument is not of the expected type
        RuntimeError:
            - When both company and companies arguments are set
            - When API request returns unexpected error

        Examples
        --------
        Get data for a single company using the code type 'isin'
            >>> taxonomy = Taxonomy()
            >>> single_company_data = taxonomy.get_company('isin', company_code='ABCNMST00394')
            >>> print(single_company_data)
                         id  fcode           common_name
            0  ABCNMST00394  ABCYT  Systemy Company S.A.

        Get data for multiple companies sugin the code type 'isin'
            >>> taxonomy = Taxonomy()
            >>> multiple_companies_data = taxonomy.get_company('isin', company_codes=['ABC3E53433100', 'XYZ233341067', 'MN943181045'])
            >>> print(multiple_companies_data)
                         id   fcode      common_name
            0  ABC3E5343310  MCABST  M**************
            1  XYZ233341067   AXYZC    A************
            2  MN9431810453     MNM     M***********

        """
        if type(company_codes) is str:
            return self.get_single_company(code_type, company_codes)
        elif type(company_codes) is list:
            return self.get_multiple_companies(code_type, company_codes)
        else:
            raise ValueError('company_codes must be a string or a list')



from .. import (UserKey, factiva_logger, get_factiva_logger, req,
                          tools)
from ..const import (API_COMPANIES_IDENTIFIER_TYPE, API_HOST,
                                API_SNAPSHOTS_COMPANIES_BASEPATH,
                                API_SNAPSHOTS_COMPANIES_PIT,
                                API_SNAPSHOTS_TAXONOMY_BASEPATH,
                                DOWNLOAD_DEFAULT_FOLDER,
                                TICKER_COMPANY_IDENTIFIER)


class Company():
    """Class that represents the company available within the Snapshots API.
    
    Parameters
    ----------
    user_key : str or UserKey
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    
    Examples
    --------
    Creating a company instance providing the user key
        >>> c = Company(u='abcd1234abcd1234abcd1234abcd1234')

    Creating a company instance with an existing UserKey instance
        >>> u = UserKey('abcd1234abcd1234abcd1234abcd1234', True)
        >>> c = Company(user_key=u)
    """

    __API_ENDPOINT_TAXONOMY = f'{API_HOST}{API_SNAPSHOTS_TAXONOMY_BASEPATH}'
    __API_ENDPOINT_COMPANY = f'{API_HOST}{API_SNAPSHOTS_COMPANIES_BASEPATH}'
    __TICKER_COMPANY_IDENTIFIER_NAME = 'ticker_exchange'

    user_key=None
    
    def __init__(self, user_key=None):
        """Class initializar"""
        self.user_key = UserKey.create_user_key(user_key, True)
        self.log= get_factiva_logger()

    @factiva_logger
    def validate_point_time_request(self, identifier):
        """Validate if the user is allowes to perform company operation and if the identifier given is valid
        
        Parameters
        ----------
        identifier : str
            A company identifier type
        
        Raises
        ------
        ValueError: When the user is not allowed to permorm this operation
        ValueError: When the identifier requested is not valid
        """

        if (not len(self.user_key.enabled_company_identifiers)):
            raise ValueError('User is not allowed to perform this operation')

        tools.validate_field_options(identifier, API_COMPANIES_IDENTIFIER_TYPE)

        if (identifier == TICKER_COMPANY_IDENTIFIER):
            identifier = self.__TICKER_COMPANY_IDENTIFIER_NAME

        identifier_description = list(
            filter(lambda company: company['name'] == identifier,
                   self.user_key.enabled_company_identifiers))
        if (not len(identifier_description)):
            raise ValueError('User is not allowed to perform this operation')

    @factiva_logger
    def point_in_time_download_all(self,
                                   identifier,
                                   file_name,
                                   file_format,
                                   to_save_path=None,
                                   add_timestamp=False) -> str:
        """Returns a file with the historical and current identifiers for each category and news coded companies.

        Parameters
        ----------
        identifier : str
            A company identifier type
        file_name : str
            Name to be used as local filename
        file_format : str
            Format of the file
        to_save_path : str, optional
            Path to be used to store the file
        add_timestamp : bool, optional
            Flag to determine if include timestamp info at the filename

        Returns
        -------
        str:
            Dowloaded file path

        Raises
        ------
        ValueError: When the user is not allowed to permorm this operation
        ValueError: When the identifier requested is not valid
        ValueError: When the format file requested is not valid
        """

        self.validate_point_time_request(identifier)
        if (to_save_path is None):
            to_save_path = DOWNLOAD_DEFAULT_FOLDER

        headers_dict = {'user-key': self.user_key.key}
        endpoint = f'{self.__API_ENDPOINT_TAXONOMY}{API_SNAPSHOTS_COMPANIES_PIT}/{identifier}/{file_format}'

        local_file_name = req.download_file(endpoint, headers_dict, file_name,
                                            file_format, to_save_path,
                                            add_timestamp)
        return local_file_name

    @factiva_logger
    def point_in_time_query(self, identifier, value) -> dict:
        """Returns the resolved Factiva code and date ranges when the instrument from the identifier, was valid.
        
        Parameters
        ----------
        identifier : str
            A company identifier type
        value : str
            Identifier value
        
        Returns
        -------
        dict:
            Factiva code and date ranges from a company
        
        Raises
        ------
        ValueError: When the user is not allowed to permorm this operation
        ValueError: When the identifier requested is not valid
        """

        self.validate_point_time_request(identifier)
        headers_dict = {'user-key': self.user_key.key}
        endpoint = f'{self.__API_ENDPOINT_COMPANY}{API_SNAPSHOTS_COMPANIES_PIT}/{identifier}/{value}'

        response = req.api_send_request(endpoint_url=endpoint,
                                        headers=headers_dict)
        if response.status_code == 200:
            response = response.json()
            return response
        else:
            raise RuntimeError(
                '''Unexpected HTTP Response from API while checking for limits'''
            )



"""
Factiva Analytics package constants
"""
from __future__ import absolute_import, division, print_function

import os

from ..tools import load_environment_value
from .errors import *

"""Define library's constant literals."""
API_HOST = 'https://api.dowjones.com'
API_ACCOUNT_OAUTH2_HOST = 'https://accounts.dowjones.com/oauth2/v1/token'
API_LATEST_VERSION = "2.0"

# UserKey
API_ACCOUNT_BASEPATH = '/alpha/accounts'
API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH = '/accounts/streaming-credentials'

# Dynamic Prefixes
ALPHA_BASEPATH = '/alpha'
DNA_BASEPATH = '/dna'  # Deprecated

# Snapshots
API_SNAPSHOTS_BASEPATH = '/alpha/extractions/documents'
API_EXPLAIN_SUFFIX = '/_explain'
API_ANALYTICS_BASEPATH = '/alpha/analytics'
API_EXTRACTIONS_BASEPATH = '/alpha/extractions'
API_EXTRACTIONS_SAMPLES_SUFFIX = '/samples'
API_DEFAULT_EXTRACTION_TYPE = "documents"
API_SAMPLES_EXTRACTION_TYPE = "samples"

API_SNAPSHOTS_TAXONOMY_BASEPATH = '/alpha/taxonomies'
API_SNAPSHOTS_COMPANIES_BASEPATH = '/alpha/companies'
API_SNAPSHOTS_COMPANY_IDENTIFIERS_BASEPATH = '/alpha/companies/identifiers'
API_SNAPSHOTS_COMPANIES_PIT = '/pit'
CUSIP_COMPANY_IDENTIFIER = 'cusip'
ISIN_COMPANY_IDENTIFIER = 'isin'
SEDOL_COMPANY_IDENTIFIER = 'sedol'
TICKER_COMPANY_IDENTIFIER = 'ticker'
API_COMPANIES_IDENTIFIER_TYPE = [
    CUSIP_COMPANY_IDENTIFIER, ISIN_COMPANY_IDENTIFIER,
    SEDOL_COMPANY_IDENTIFIER, TICKER_COMPANY_IDENTIFIER
]

# ANALYTICS
API_AVRO_FORMAT = 'avro'
API_CSV_FORMAT = 'csv'
API_JSON_FORMAT = 'json'
API_EXTRACTION_FILE_FORMATS = [
    API_AVRO_FORMAT, API_JSON_FORMAT, API_CSV_FORMAT
]

API_DAY_PERIOD = 'DAY'
API_MONTH_PERIOD = 'MONTH'
API_YEAR_PERIOD = 'YEAR'
API_DATETIME_PERIODS = [API_DAY_PERIOD, API_MONTH_PERIOD, API_YEAR_PERIOD]

API_PUBLICATION_DATETIME_FIELD = 'publication_datetime'
API_MODIFICATION_DATETIME_FIELD = 'modification_datetime'
API_INGESTION_DATETIME_FIELD = 'ingestion_datetime'
API_DATETIME_FIELDS = [
    API_PUBLICATION_DATETIME_FIELD, API_MODIFICATION_DATETIME_FIELD,
    API_INGESTION_DATETIME_FIELD
]
API_GROUP_DIMENSIONS_FIELDS = [
    'source_code', 'subject_codes', 'region_codes', 'industry_codes',
    'company_codes', 'person_codes', 'company_codes_about',
    'company_codes_relevance', 'company_codes_cusip', 'company_codes_isin',
    'company_codes_sedol', 'company_codes_ticker', 'company_codes_about_cusip',
    'company_codes_about_isin', 'company_codes_about_sedol',
    'company_codes_about_ticker', 'company_codes_relevance_cusip',
    'company_codes_relevance_isin', 'company_codes_relevance_sedol',
    'company_codes_relevance_ticker'
]

# Streams
API_STREAMS_BASEPATH = '/alpha/streams'
DOC_COUNT_EXCEEDED = "DOC_COUNT_EXCEEDED"
CHECK_EXCEEDED_WAIT_SPACING = 300
PUBSUB_MESSAGES_WAIT_SPACING = 10

# API STATES
API_JOB_CREATED_STATE = 'JOB_CREATED'
API_JOB_QUEUED_STATE = 'JOB_QUEUED'
API_JOB_PENDING_STATE = 'JOB_STATE_PENDING'
API_JOB_VALIDATING_STATE = 'JOB_VALIDATING'
API_JOB_STATE_VALIDATING = 'JOB_STATE_VALIDATING'
API_JOB_RUNNING_STATE = 'JOB_STATE_RUNNING'
API_JOB_DONE_STATE = 'JOB_STATE_DONE'
API_JOB_FAILED_STATE = 'JOB_STATE_FAILED'
API_JOB_CANCELLED_STATE = 'JOB_STATE_CANCELLED'

API_JOB_EXPECTED_STATES = [
    API_JOB_CREATED_STATE, API_JOB_QUEUED_STATE, API_JOB_PENDING_STATE,
    API_JOB_VALIDATING_STATE, API_JOB_STATE_VALIDATING, API_JOB_RUNNING_STATE,
    API_JOB_DONE_STATE, API_JOB_FAILED_STATE, API_JOB_CANCELLED_STATE
]

API_JOB_ACTIVE_WAIT_SPACING = 10

# SNAPSHOT FILES
SNAPSHOT_FILE_STATS_FIELDS = [
    'an', 'company_codes', 'company_codes_about', 'company_codes_occur',
    'industry_codes', 'ingestion_datetime', 'language_code',
    'modification_datetime', 'publication_datetime', 'publisher_name',
    'region_codes', 'region_of_origin', 'source_code', 'source_name',
    'subject_codes', 'title', 'word_count'
]

SNAPSHOT_FILE_DELETE_FIELDS = [
    'art', 'credit', 'document_type', 'publication_date', 'modfication_date'
]  # publication_date and modification_date are deprecated

# Files options
DOWNLOAD_DEFAULT_FOLDER = load_environment_value(
    "DOWNLOAD_FILES_DIR", os.path.join(os.getcwd(), 'downloads'))
LISTENER_FILES_DEFAULT_FOLDER = load_environment_value(
    'STREAM_FILES_DIR', os.path.join(os.getcwd(), 'listener'))
LOGS_DEFAULT_FOLDER = load_environment_value(
    'LOG_FILES_DIR', os.path.join(os.getcwd(), 'logs'))

#TIMESTAMP
TIMESTAMP_FIELDS = [
    "ingestion_datetime", "modification_date", "modification_datetime",
    "publication_date", "publication_datetime"
]

#Multivalue
MULTIVALUE_FIELDS_COMMA = [
    'company_codes', 'subject_codes', 'region_codes', 'industry_codes',
    'person_codes', 'currency_codes', 'market_index_codes',
    'company_codes_about', 'company_codes_association',
    'company_codes_lineage', 'company_codes_occur', 'company_codes_relevance',
    'company_codes_about_cusip', 'company_codes_association_cusip',
    'company_codes_lineage_cusip', 'company_codes_occur_cusip',
    'company_codes_relevance_cusip', 'company_codes_cusip',
    'company_codes_about_isin', 'company_codes_association_isin',
    'company_codes_lineage_isin', 'company_codes_occur_isin',
    'company_codes_relevance_isin', 'company_codes_isin',
    'company_codes_about_ticker_exchange',
    'company_codes_association_ticker_exchange',
    'company_codes_lineage_ticker_exchange',
    'company_codes_occur_ticker_exchange',
    'company_codes_relevance_ticker_exchange', 'company_codes_ticker_exchange',
    'company_codes_about_sedol', 'company_codes_association_sedol',
    'company_codes_lineage_sedol', 'company_codes_occur_sedol',
    'company_codes_relevance_sedol', 'company_codes_sedol'
]

MULTIVALUE_FIELDS_SPACE = ["region_of_origin"]

ADD_ACTION = 'add'
REP_ACTION = 'rep'
DEL_ACTION = 'del'
ERR_ACTION = 'error'
ALLOWED_ACTIONS = [ADD_ACTION, REP_ACTION, DEL_ACTION]
ACTION_CONSOLE_INDICATOR = {}
ACTION_CONSOLE_INDICATOR[ADD_ACTION] = '.'
ACTION_CONSOLE_INDICATOR[REP_ACTION] = ':'
ACTION_CONSOLE_INDICATOR[DEL_ACTION] = '&'
ACTION_CONSOLE_INDICATOR[ERR_ACTION] = '!'



COLUMN_NAMES = [
    "action",
    "an",
    "art",
    "body",
    "byline",
    "company_codes",
    "company_codes_about",
    "company_codes_about_ticker_exchange",
    "company_codes_association",
    "company_codes_association_ticker_exchange",
    "company_codes_lineage",
    "company_codes_lineage_ticker_exchange",
    "company_codes_occur",
    "company_codes_occur_ticker_exchange",
    "company_codes_relevance",
    "company_codes_relevance_ticker_exchange",
    "company_codes_ticker_exchange",
    "copyright",
    "credit",
    "currency_codes",
    "delivery_datetime",
    "document_type",
    "industry_codes",
    "ingestion_datetime",
    "language_code",
    "market_index_codes",
    "modification_date",
    "modification_datetime",
    "person_codes",
    "publication_date",
    "publication_datetime",
    "publisher_name",
    "region_codes",
    "region_of_origin",
    "section",
    "snippet",
    "source_code",
    "source_name",
    "subject_codes",
    "title",
    "word_count",
]


def format_message_to_response_schema(msg):
    new = {k: v for (k, v) in msg.items() if k in COLUMN_NAMES}
    return new



"""Define common errors in the library."""
UNEXPECTED_HTTP_ERROR = RuntimeError('API Request returned an unexpected HTTP status')
UNDEFINED_STREAM_ID_ERROR = ValueError('Undefined stream id')
INVALID_SUBSCRIPTION_ID_ERROR = ValueError('Invalid subscription id')
UNDEFINED_SUBSCRIPTION_ERROR = ValueError('No subscription specified. You must specify the subscription ID')



"""
    Module to handle the API and other requests
"""
import json
import os
from datetime import datetime
import requests
from .. import const, tools
from ..log import factiva_logger, get_factiva_logger

log = get_factiva_logger()

def send_get_request(endpoint_url=const.API_HOST,
                     headers=None,
                     qs_params=None,
                     stream=False):
    """Send get request."""
    if (qs_params is not None) and (not isinstance(qs_params, dict)):
        raise ValueError('Unexpected qs_params value')
    return requests.get(endpoint_url,
                        headers=headers,
                        params=qs_params,
                        stream=stream)


def send_post_request(endpoint_url=const.API_HOST, headers=None, payload=None):
    """Send post request."""
    if payload is not None:
        if isinstance(payload, dict):
            payload_str = json.dumps(payload)
        elif isinstance(payload, str):
            payload_str = payload
        else:
            raise ValueError('Unexpected payload value')
        return requests.post(endpoint_url, headers=headers, data=payload_str)

    return requests.post(endpoint_url, headers=headers)

@factiva_logger()
def api_send_request(method='GET',
                     endpoint_url=const.API_HOST,
                     headers=None,
                     payload=None,
                     qs_params=None,
                     stream=False):
    """Send a generic request to a certain API end point."""
    if headers is None:
        raise ValueError('Heders for Factiva requests cannot be empty')

    if not isinstance(headers, dict):
        raise ValueError('Unexpected headers value')

    try:
        if method == 'GET':
            response = send_get_request(endpoint_url=endpoint_url,
                                        headers=headers,
                                        qs_params=qs_params,
                                        stream=stream)

        elif method == 'POST':
            response = send_post_request(endpoint_url=endpoint_url,
                                         headers=headers,
                                         payload=payload)

        elif method == 'DELETE':
            response = requests.delete(endpoint_url, headers=headers)

        else:
            raise ValueError('Unexpected method value')

    except Exception:
        raise RuntimeError('API Request failed. Unspecified Error.')

    return response


@factiva_logger()
def download_file(file_url,
                  headers,
                  file_name,
                  file_extension,
                  to_save_path,
                  add_timestamp=False) -> str:
    """Download a file on a specific path.
    
    Parameters
    ----------
    file_url : str
        URL of the file to be downloaded.
    headers : dict
        Auth headers
    file_name : str
        Name to be used as local filename
    file_extension : str
        Extension of the file
    to_save_path : str
        Path to be used to store the file
    add_timestamp : bool, optional
        Flag to determine if include timestamp info at the filename
    
    Returns
    -------
    str
        Dowloaded file path
    """

    tools.validate_field_options(file_extension,
                                 const.API_EXTRACTION_FILE_FORMATS)

    tools.create_path_if_not_exist(to_save_path)

    if add_timestamp:
        file_name = f'{file_name}-{datetime.now()}'

    response = send_get_request(endpoint_url=file_url,
                                headers=headers,
                                stream=True)

    local_file_name = os.path.join(to_save_path,
                                   f'{file_name}.{file_extension}')
    with open(local_file_name, 'wb') as f:
        f.write(response.content)

    return local_file_name



import functools
import logging
import os
import sys
from inspect import getframeinfo, stack
from .const import LOGS_DEFAULT_FOLDER
import datetime

class CustomFormatter(logging.Formatter):
    """ Custom Formatter does these 2 things:
    1. Overrides 'funcName' with the value of 'func_name_override', if it exists.
    2. Overrides 'filename' with the value of 'file_name_override', if it exists.
    """

    def format(self, record):
        if hasattr(record, 'func_name_override'):
            record.funcName = record.func_name_override
        if hasattr(record, 'file_name_override'):
            record.filename = record.file_name_override
        return super(CustomFormatter, self).format(record)


def get_factiva_logger():
    if not os.path.exists(LOGS_DEFAULT_FOLDER):
        os.mkdir(LOGS_DEFAULT_FOLDER)
    logger = logging.Logger(__name__)
    logger.setLevel(logging.DEBUG)
    file_name = f'factiva-{datetime.datetime.now().strftime("%d-%m-%Y-%H")}'
    handler = logging.FileHandler(
        "{0}/{1}.log".format(LOGS_DEFAULT_FOLDER, file_name), 'a+')
    handler.setFormatter(
        CustomFormatter(
            '%(asctime)s - %(levelname)-10s - %(filename)s - %(funcName)s - %(message)s'
        ))
    logger.addHandler(handler)
    return logger


def factiva_logger(_func=None):
    def log_decorator_info(func):
        @functools.wraps(func)
        def factiva_log(self=None, *args, **kwargs):
            logger_obj = get_factiva_logger()
            # args_passed_in_function = [repr(a) for a in args]
            # kwargs_passed_in_function = [
            #     f"{k}={v!r}" for k, v in kwargs.items()
            # ]
            # formatted_arguments = ", ".join(args_passed_in_function +
            #                                 kwargs_passed_in_function)
            py_file_caller = getframeinfo(stack()[1][0])
            extra_args = {
                'func_name_override': func.__name__,
                'file_name_override': os.path.basename(py_file_caller.filename)
            }
            logger_obj.debug(
                "Begin function",
                extra=extra_args)
            try:
                if(self is None):
                    value = func(*args, **kwargs)
                else:
                    value = func(self, *args, **kwargs)
                logger_obj.debug("End function")
            except:
                logger_obj.error("Exception: {}".format(str(sys.exc_info()[1])))
                raise
            return value

        return factiva_log

    if _func is None:
        return log_decorator_info
    else:
        return log_decorator_info(_func)



"""Factiva Core UserKey Class."""
import json
import pandas as pd
from .. import const
from ..log import factiva_logger, get_factiva_logger
from ..req import api_send_request
from ..tools import flatten_dict, load_environment_value, mask_string


class UserKey:
    """Class that represents an API user. This entity is identifiable by a User Key.

    Parameters
    ----------
    key : str
        String containing the 32-character long APi Key. If not provided, the
        constructor will try to obtain its value from the FACTIVA_USERKEY
        environment variable.
    stats : bool, optional (Default: False)
        Indicates if user data has to be pulled from the server. This operation
        fills account detail properties along with maximum, used and remaining
        values. It may take several seconds to complete.

    Examples
    --------
    Creating a new UserKey instance providing the key string explicitly and requesting to retrieve the latest account details:
        >>> u = UserKey('abcd1234abcd1234abcd1234abcd1234', stats=True)
        >>> print(u)
            <class 'factiva.core.userkey.UserKey'>
            |-key = ****************************1234
            |-cloud_token = **Not Fetched**
            |-account_name = AccName1234
            |-account_type = account_with_contract_limits
            |-active_products = DNA
            |-max_allowed_concurrent_extractions = 5
            |-max_allowed_extracted_documents = 200,000
            |-max_allowed_extractions = 3
            |-currently_running_extractions = 0
            |-total_downloaded_bytes = 7,253,890
            |-total_extracted_documents = 2,515
            |-total_extractions = 1
            |-total_stream_instances = 4
            |-total_stream_subscriptions = 1
            |-enabled_company_identifiers = [{'id': 4, 'name': 'isin'}, {'id': 3, 'name': 'cusip'}, {'id': 1, 'name': 'sedol'}, {'id': 5, 'name': 'ticker_exchange'}]
            |-remaining_documents = 197,485
            |-remaining_extractions = 2

    Creating a new instance taking the key value from the environment varaible FACTIVA_USERKEY, and not requesting account statistics (default).
        >>> u = UserKey()
        >>> print(u)
            <class 'factiva.core.userkey.UserKey'>
            |-key = ****************************1234
            |-cloud_token = **Not Fetched**
            |-account_name =
            |-account_type =
            |-active_products =
            |-max_allowed_concurrent_extractions = 0
            |-max_allowed_extracted_documents = 0
            |-max_allowed_extractions = 0
            |-currently_running_extractions = 0
            |-total_downloaded_bytes = 0
            |-total_extracted_documents = 0
            |-total_extractions = 0
            |-total_stream_instances = 0
            |-total_stream_subscriptions = 0
            |-enabled_company_identifiers = []
            |-remaining_documents = 0
            |-remaining_extractions = 0

    """
    # pylint: disable=too-many-instance-attributes
    # Twelve is reasonable in this case.

    __API_ENDPOINT_BASEURL = f'{const.API_HOST}{const.API_ACCOUNT_BASEPATH}/'
    __API_CLOUD_TOKEN_URL = f'{const.API_HOST}{const.ALPHA_BASEPATH}{const.API_ACCOUNT_STREAM_CREDENTIALS_BASEPATH}'

    key = ''
    cloud_token = {}
    account_name = ''
    account_type = ''
    active_products = ''
    max_allowed_concurrent_extractions = 0
    max_allowed_extracted_documents = 0
    max_allowed_extractions = 0
    currently_running_extractions = 0
    total_downloaded_bytes = 0
    total_extracted_documents = 0
    total_extractions = 0
    total_stream_instances = 0
    total_stream_subscriptions = 0
    enabled_company_identifiers = []

    def __init__(self, key=None, stats=False):
        """Construct the instance of the class."""
        self.log = get_factiva_logger()
        if key is None:
            try:
                key = load_environment_value('FACTIVA_USERKEY')
            except Exception as error:
                raise ValueError(
                    'Factiva user key not provided and environment variable FACTIVA_USERKEY not set.'
                ) from error

        if len(key) != 32:
            raise ValueError('Factiva User-Key has the wrong length')

        self.key = key
        self.cloud_token = {}

        if stats is True:
            self.get_stats()
        else:
            self.account_name = ''
            self.account_type = ''
            self.active_products = ''
            self.max_allowed_concurrent_extractions = 0
            self.max_allowed_extracted_documents = 0
            self.max_allowed_extractions = 0
            self.currently_running_extractions = 0
            self.total_downloaded_bytes = 0
            self.total_extracted_documents = 0
            self.total_extractions = 0
            self.total_stream_instances = 0
            self.total_stream_subscriptions = 0
            self.enabled_company_identifiers = []


    @property
    def remaining_extractions(self):
        """Account remaining extractions."""
        return self.max_allowed_extractions - self.total_extractions


    @property
    def remaining_documents(self):
        """Account remaining documents."""
        return self.max_allowed_extracted_documents - self.total_extracted_documents

    @property
    def extractions_done(self):
        """Number of executed extractions"""
        return self.get_extractions()

    @property
    def streams_running(self):
        """Number of currently running Streaming Instances"""
        return self.get_streams()

    @factiva_logger()
    def get_stats(self) -> bool:
        """Request the account details to the Factiva Account API Endpoint.

        This operation can take several seconds to complete.

        Returns
        -------
        True if the operation was completed successfully. All returned values
        are assigned to the object's properties directly.

        Examples
        --------
        >>> u = UserKey('abcd1234abcd1234abcd1234abcd1234')
        >>> print(u)
            <class 'factiva.core.userkey.UserKey'>
            |-key = ****************************1234
            |-cloud_token = **Not Fetched**
            |-account_name =
            |-account_type =
            |-active_products =
            |-max_allowed_concurrent_extractions = 0
            |-max_allowed_extracted_documents = 0
            |-max_allowed_extractions = 0
            |-currently_running_extractions = 0
            |-total_downloaded_bytes = 0
            |-total_extracted_documents = 0
            |-total_extractions = 0
            |-total_stream_instances = 0
            |-total_stream_subscriptions = 0
            |-enabled_company_identifiers = []
            |-remaining_documents = 0
            |-remaining_extractions = 0
        >>> u.get_stats()
        >>> print(u)
            <class 'factiva.core.userkey.UserKey'>
            |-key = ****************************1234
            |-cloud_token = **Not Fetched**
            |-account_name = AccName1234
            |-account_type = account_with_contract_limits
            |-active_products = DNA
            |-max_allowed_concurrent_extractions = 5
            |-max_allowed_extracted_documents = 200,000
            |-max_allowed_extractions = 3
            |-currently_running_extractions = 0
            |-total_downloaded_bytes = 7,253,890
            |-total_extracted_documents = 2,515
            |-total_extractions = 1
            |-total_stream_instances = 4
            |-total_stream_subscriptions = 1
            |-enabled_company_identifiers = [{'id': 4, 'name': 'isin'}, {'id': 3, 'name': 'cusip'}, {'id': 1, 'name': 'sedol'}, {'id': 5, 'name': 'ticker_exchange'}]
            |-remaining_documents = 197,485
            |-remaining_extractions = 2

        """
        account_endpoint = f'{self.__API_ENDPOINT_BASEURL}{self.key}'
        req_head = {'user-key': self.key}
        resp = api_send_request(method='GET', endpoint_url=account_endpoint, headers=req_head)
        if resp.status_code == 200:
            try:
                resp_obj = json.loads(resp.text)
                self.account_name = resp_obj['data']['attributes']['name']
                self.account_type = resp_obj['data']['type']
                self.active_products = resp_obj['data']['attributes']['products']
                self.max_allowed_concurrent_extractions = resp_obj['data']['attributes']['max_allowed_concurrent_extracts']
                self.max_allowed_extracted_documents = resp_obj['data']['attributes']['max_allowed_document_extracts']
                self.max_allowed_extractions = resp_obj['data']['attributes']['max_allowed_extracts']
                self.currently_running_extractions = resp_obj['data']['attributes']['cnt_curr_ext']
                self.total_downloaded_bytes = resp_obj['data']['attributes']['current_downloaded_amount']
                self.total_extracted_documents = resp_obj['data']['attributes']['tot_document_extracts']
                self.total_extractions = resp_obj['data']['attributes']['tot_extracts']
                self.total_stream_instances = resp_obj['data']['attributes']['tot_topics']
                self.total_stream_subscriptions = resp_obj['data']['attributes']['tot_subscriptions']
                self.enabled_company_identifiers = resp_obj['data']['attributes']['enabled_company_identifiers']
            except Exception as error:
                raise AttributeError('Unexpected Account Information API Response.') from error
        elif resp.status_code == 403:
            raise ValueError('Factiva User-Key does not exist or inactive.')
        else:
            raise RuntimeError('Unexpected Account Information API Error')
        return True

    @factiva_logger()
    def get_cloud_token(self) -> bool:
        """
        Request a cloud token to the API and saves its value
        in the cloud_token property

        Returns
        -------
        True if the operation was completed successfully. Returned value
        is assigned to the cloud_token property.

        Raises
        ------
        ValueError: When the credentials are not valid
        RuntimeError: When API request returns unexpected error

        """
        req_head = {'user-key': self.key}
        response = api_send_request(
            method="GET",
            endpoint_url=f'{self.__API_CLOUD_TOKEN_URL}',
            headers=req_head
        )

        if response.status_code == 401:
            message = '''
                Extraction API authentication failed for given
                credentials header:{}
                '''.format(req_head)
            raise RuntimeError(message)
        try:
            streaming_credentials_string = response.json()['data']['attributes']['streaming_credentials']
        except TypeError as type_error:
            raise ValueError('Unable to get a cloud token for the given key. This account might have limited access.') from type_error

        self.cloud_token = json.loads(streaming_credentials_string)
        return True

    @factiva_logger()
    def get_extractions(self, updates=False) -> pd.DataFrame:
        """Request a list of the extractions of the account.

        Parameters
        ----------
        updates : bool
            Flag that indicates whether the retrieved list should include (True)
            or not (False) Snapshot Update calls.

        Returns
        -------
        Dataframe containing the list of historical extractions for the account.

        Raises
        ------
        - ValueError when the API Key provided is not valid
        - RuntimeError when the API returns an unexpected error

        """
        endpoint = f'{const.API_HOST}{const.API_EXTRACTIONS_BASEPATH}'

        headers_dict = {'user-key': self.key}

        response = api_send_request(method='GET', endpoint_url=endpoint, headers=headers_dict)

        if response.status_code != 200:
            if response.status_code == 403:
                raise ValueError('Factiva API-Key does not exist or inactive.')

            raise RuntimeError(f'Unexpected API Error with message: {response.text}')

        response_data = response.json()

        extraction_df = pd.DataFrame([flatten_dict(extraction) for extraction in response_data['data']])
        extraction_df.rename(columns={'id': 'object_id'}, inplace=True)
        ids_df = extraction_df['object_id'].str.split('-', expand=True)

        if ids_df.shape[1] >= 5:
            extraction_df['snapshot_sid'] = ids_df[4]
        else:
            extraction_df['snapshot_sid'] = None

        if ids_df.shape[1] >= 7:
            extraction_df['update_id'] = ids_df[6]
        else:
            extraction_df['update_id'] = None

        extraction_df.drop(['self', 'type'], axis=1, inplace=True)

        if not updates:
            extraction_df = extraction_df.loc[extraction_df.update_id.isnull()]

        return extraction_df


    def show_extractions(self, updates=False):
        """Shows a list of the extractions of the account.

        Parameters
        ----------
        updates : bool
            Flag that indicates whether the displayed list should include (True)
            or not (False) Snapshot Update calls.

        Returns
        -------
        Dataframe containing the list of historical extractions for the account.

        Raises
        ------
        - ValueError when the API Key provided is not valid
        - RuntimeError when the API returns an unexpected error

        """
        extractions = self.get_extractions(updates=updates)
        print(extractions.loc[:, extractions.columns != 'object_id'])


    @factiva_logger()
    def get_streams(self, running=True) -> pd.DataFrame:
        """Obtain streams from a given user.

        Function which returns the streams a given user with
        its respective key using the default stream url

        Parameters
        ----------
        running : bool
            Flag that indicates whether the retrieved list should be restricted
            to only running streams (True) or also include cancelled and failed
            ones (False).

        Returns
        -------
        DataFrame -> DataFrame with the list of historical extractions

        Raises
        ------
        AttributeError:
            When is not possible to parse the data as json or dataframe
        ValueError:
            When API key is not valid
        RuntimeError:
            When API request returns unexpected error

        """
        request_headers = {'user-key': self.key}
        response = api_send_request(
            method="GET",
            endpoint_url=f'{const.API_HOST}{const.API_STREAMS_BASEPATH}',
            headers=request_headers
        )
        if response.status_code == 200:
            try:
                def extract_subscriptions(subscription):
                    # Fixed issue#12 (https://github.com/dowjones/factiva-core-python/issues/12)
                    id_list = []
                    for i in subscription:
                        s_idp = i['id'].split('-')
                        s_id = f"{s_idp[-3]}-{s_idp[-2]}-{s_idp[-1]}"
                        id_list.append(s_id)
                    return id_list

                response_data = response.json()
                stream_df = pd.DataFrame([flatten_dict(extraction) for extraction in response_data['data']])
                stream_df.rename(columns={'id': 'object_id'}, inplace=True)
                ids_df = stream_df['object_id'].str.split('-', expand=True)
                stream_df['stream_id'] = ids_df[4]
                stream_df['stream_type'] = ids_df[2]
                stream_df['subscriptions'] = stream_df['data'].apply(extract_subscriptions)
                stream_df['n_subscriptions'] = stream_df['subscriptions'].str.len()
                stream_df.drop(['self', 'type', 'data'], axis=1, inplace=True)

                if running:
                    stream_df = stream_df.loc[stream_df.job_status == const.API_JOB_RUNNING_STATE]

                return stream_df
            except Exception as error:
                raise AttributeError('Unexpected Get Streams API Response.') from error
        elif response.status_code == 403:
            raise ValueError('Factiva API-Key does not exist or inactive.')
        else:
            raise RuntimeError('Unexpected Get Streams API Error')


    def show_streams(self, running=True):
        """Shows the list of streams for a given user.

        This function runs the existing function get_streams and
        prints a user-friendly table with stream details.

        Parameters
        ----------
        running : bool
            Flag that indicates whether the displayed list should be restricted
            to only running streams (True) or also include cancelled and failed
            ones (False).

        Returns
        -------
        None

        Raises
        ------
        AttributeError:
            When is not possible to parse the data as json or dataframe
        ValueError:
            When API key is not valid
        RuntimeError:
            When API request returns unexpected error

        """
        account_streams = self.get_streams(running=running)
        print(account_streams.loc[:, account_streams.columns != 'object_id'])


    def is_active(self) -> bool:
        request_headers = {'user-key': self.key}
        response = api_send_request(
            method="GET",
            endpoint_url=f'{const.API_HOST}{const.API_SNAPSHOTS_TAXONOMY_BASEPATH}',
            headers=request_headers
        )
        if response.status_code == 200:
            return True
        else:
            return False


    def __print_property__(self, property_value) -> str:
        if isinstance(property_value, int):
            pval = f'{property_value:,d}'
        else:
            pval = property_value
        return pval


    def __repr__(self):
        """Return a string representation of the object."""
        return self.__str__()


    def __str__(self, detailed=True, prefix='  |-', root_prefix=''):
        # TODO: Improve the output for enabled_company_identifiers
        pprop = self.__dict__.copy()
        del pprop['key']
        del pprop['cloud_token']
        masked_key = mask_string(self.key)
        if self.cloud_token == {}:
            masked_token = '**Not Fetched**'
        else:
            masked_token = mask_string(self.cloud_token['private_key'][58:92], 12)

        ret_val = f'{root_prefix}{str(self.__class__)}\n'
        ret_val += f'{prefix}key = {masked_key}\n'
        ret_val += f'{prefix}cloud_token = {masked_token}\n'
        if detailed:
            ret_val += '\n'.join((f'{prefix}{item} = {self.__print_property__(pprop[item])}' for item in pprop))
            ret_val += f'\n{prefix}remaining_documents = {self.__print_property__(self.remaining_documents)}\n'
            ret_val += f'{prefix}remaining_extractions = {self.__print_property__(self.remaining_extractions)}\n'
        else:
            ret_val += f'{prefix}...'
        return ret_val

    @staticmethod
    def create_user_key(key=None, stats=False):
        """Determine the way to initialize an api key user according to the type of parameter provided.

        Parameters
        ----------
        api_user : None, str, UserKey
                Source to create a UserKey instance
        stats : boolean, optional (Default: False)
                Indicates if user data has to be pulled from the server

        Returns
        -------
        UserKey instance accordingly:
            - When None is passed, UserKey instance using credentials from ENV variables
            - When str is passed, UserKey instance using the provided parameter as credentials
            - When UserKey is passed, it returns the same instance

        Raises
        ------
            RuntimeError: When an UserKey instance cannot be created using the provided parameters

        """
        if isinstance(key, UserKey):
            return key

        if isinstance(key, str):
            try:
                return UserKey(key, stats=stats)
            except Exception as error:
                raise RuntimeError("User cannot be obtained from the provided key.") from error

        if key is None:
            try:
                return UserKey(stats=stats)
            except Exception as error:
                raise RuntimeError("User cannot be obtained from ENV variables") from error

        raise RuntimeError("Unexpected api_user value")



"""Define basic dictionaries of Hierarchies adn Taxonomies."""
import os
import pandas as pd
import numpy as np

rootdir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
datadir = os.path.join(rootdir, 'dicts')
ind_hrchy_path = os.path.join(datadir, 'industries-hrchy.csv')
reg_hrchy_path = os.path.join(datadir, 'regions-hrchy.csv')
countries_path = os.path.join(datadir, 'factiva-countries.csv')


def industries_hierarchy() -> pd.DataFrame:
    """Read the Dow Jones Industry hierarchy CSV file.

    Reads the Dow Jones Industry hierarchy CSV file and returns
    its content as a Pandas DataFrame. The root node has
    the fcode `indroot` and an empty parent.

    Returns
    -------
    DataFrame : A Pandas DataFrame with the columns:
        * ind_fcode : string
            Industry Factiva Code
        * name : string
            Name of the Industry
        * parent : string
            Factiva Code of the parent Industry

    """
    ret_ind = pd.read_csv(ind_hrchy_path)
    ret_ind = ret_ind.replace(np.nan, '', regex=True)
    return ret_ind


def regions_hierarchy() -> pd.DataFrame:
    """Read the Dow Jones Regions hierarchy CSV file.

    Reads the Dow Jones Regions hierarchy CSV file and returns
    its content as a Pandas DataFrame. The root node has
    the fcode `WORLD` and an empty parent.

    Names containng an asterisk denote nodes not officially in
    the hierarchy, but that help balancing and reading the
    structure. Level balancing is quite useful specially for
    visualising data.

    Returns
    -------
    DataFrame : A Pandas DataFrame with the columns:
        * name : string
            Name of the region node
        * reg_fcode : string
            Factiva Code of the region
        * parent : string
            Factiva Code of the parent region
        * level : int
            Level number of the node

    """
    ret_reg = pd.read_csv(reg_hrchy_path)
    ret_reg = ret_reg.replace(np.nan, '', regex=True)
    return ret_reg


def countries_list() -> pd.DataFrame:
    """Read a list of official countries.

    Reads a list of official countries with several additional fields that
    are helpful in data merges. All contries have the Factiva Code along with
    other identifiers.

    Returns
    -------
    DataFrame : A Pandas DataFrame

    """
    ret_reg = pd.read_csv(countries_path)
    ret_reg['factiva_code'] = ret_reg['factiva_code'].str.lower()
    ret_reg.set_index('factiva_code', inplace=True)
    ret_reg = ret_reg.replace(np.nan, '', regex=True)
    return ret_reg
